# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionHandler do
  @moduledoc """
  This module is responsible for handling incoming packets and sending outgoing packets. It keeps track of what state the connection is in and which game should
  receive the packets.
  """
  alias Amethyst.Minecraft.Write
  require Logger

  defstruct handler: self(),
    version: 0,
    connection_state: Amethyst.ConnectionState.Handshake,
    game: nil,
    encryption_state: nil,
    decryption_state: nil,
    compression_threshold: nil,
    authenticated: false,
    position: {0, 0, 0},
    player_state: %{}

  @spec child_spec(:gen_tcp.socket()) :: Supervisor.child_spec()
  def child_spec(socket) do
    %{
      id: __MODULE__,
      start: {__MODULE__, :start, [socket]}
    }
  end

  @spec start(:gen_tcp.socket(), atom(), integer()) :: no_return()
  def start(socket) do
    {:ok, spawn(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(socket, %__MODULE__{handler: self()})
    end)}
  end

  @spec start_link(:gen_tcp.socket(), atom(), integer()) :: no_return()
  def start_link(socket) do
    {:ok, spawn_link(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(socket, %__MODULE__{handler: self()})
    end)}
  end

  @doc """
  Informs the `handler` that a packet was received and that it should handle it.
  """
  @spec packet(handler :: pid(), id :: pos_integer(), data :: binary())
  def packet(handler, id, data) do
    send(handler, {:packet, id, data})
  end

  @spec close(handler :: pid())
  def close(handler) do

  end

  @spec loop(:gen_tcp.socket(), %__MODULE__{}) :: no_return()
  defp loop(socket, state) do
    receive do
      :closed ->
        # Socket is closed, handle our side of the disconnect
        Process.exit(self(), :normal)
      {:packet, id, data} ->
        state = Enum.reduce(handle_packet(id, data, state), state, fn op, state -> op.(state) end)
        loop(socket, state)
      {:run, ops} ->
        state = Enum.reduce(ops, state, fn op, state -> op.(state) end)
        loop(socket, state)
      {:get, from, op} ->
        send(from, op.(state))
        loop(socket, state)
    end
  end

  defp chunk_pos(x, z) do
    {floor(round(x) / 16.0), floor(round(z) / 16.0)}
  end

  # x, z here is chunk position
  defp visible_chunks_from(x, z, view_distance) do
    (x - view_distance - 3 .. x + view_distance + 3) |> Enum.flat_map(fn ix ->
      (z - view_distance - 3 .. z + view_distance + 3) |> Enum.map(fn iz ->
        {ix, iz}
      end)
    end)
  end

  defp process_chunk(to, chunk, state) do
    import Amethyst.NBT.Write
    alias Amethyst.Minecraft.Write

    chunk_array = Amethyst.Game.chunk(Map.get(state, :game), chunk)
    {cx, cz} = chunk

    # TODO: Actually do heightmaps
    # TODO: Doing all this processing could be at home somewhere else, as here it's
    # not version-agnostic here
    heightmaps = compound(%{})

    data = Enum.chunk_every(chunk_array, 16, 16, 0) # 0 -> air
      |> Enum.reduce("", fn chunk_section, acc ->
        blocks_and_lights = chunk_section |> List.flatten()
        block_count = blocks_and_lights |> Enum.filter(fn {bs, _, _} -> bs != 0 end) |> length

        # Put together the palette
        unique_blocks = MapSet.new(blocks_and_lights |> Enum.map(fn {bs, _, _} -> bs end))
        min_bpe = MapSet.size(unique_blocks) |> :math.log2() |> ceil()

        paletted_container_data = case min_bpe do
          0 ->
          # SINGLE VALUED
            Write.ubyte(0) <>
            Write.varint(MapSet.to_list(unique_blocks) |> List.first()) <>
            Write.varint(0) # No data, empty pallette

          min_bpe when min_bpe in 1..8 ->
          # INDIRECT
            # Minimum bpe accepted by minecraft is 4
            bpe = max(min_bpe, 4)
            palette = MapSet.to_list(unique_blocks) |>
              Enum.with_index() |>
              Map.new(fn {v, i} -> {v, i} end)
            paletted_blocks = blocks_and_lights |>
              Enum.map(fn {bs, _, _} -> Map.get(palette, bs) end)
            paletted_data = long_aligned_bit_string_reduce(paletted_blocks, bpe,
              fn bs -> <<bs::signed-integer-big-size(bpe)>> end)
            Write.ubyte(bpe) <>
            Write.varint(map_size(palette)) <>
            (Enum.sort(palette, fn {_k1, v1}, {_k2, v2} -> v1 < v2 end) |>
              Enum.reduce("", fn {k, _v}, acc -> acc <> Write.varint(k) end)) <>
            Write.varint(floor(bit_size(paletted_data) / 64)) <>
            paletted_data

          _ ->
          # DIRECT
            data = long_aligned_bit_string_reduce(blocks_and_lights, 15,
              fn {bs, _sl, _bl} -> <<bs::signed-integer-big-size(15)>> end)
            Write.ubyte(15) <>
            Write.varint(floor(bit_size(data) / 64)) <>
            data
        end

        acc <> Write.short(block_count) <> paletted_container_data <>
        <<0::8, 0::8, 0::8>> # TODO: This should be biome data
      end)

  sky_light_sections = Enum.chunk_every(chunk_array, 16, 16, 0)
      |> Enum.map(fn chunk_section ->
        chunk_section |> List.flatten() |> Enum.map(fn {_bs, sl, _bl} -> sl end)
      end)
  non_empty_sky_light_sections = sky_light_sections |> Enum.filter(fn section -> !Enum.all?(section, &(&1 == 0)) end)
  has_sky_light = [false] ++ (sky_light_sections |> Enum.map(fn section -> !Enum.all?(section, &(&1 == 0)) end)) ++ [false]
  empty_sky_light = has_sky_light |> Enum.map(&not/1)
  sky_light_array = non_empty_sky_light_sections |> Enum.map(&(Enum.reduce(&1, <<>>, fn light, acc ->
    <<acc::bitstring, light::size(4)>>
  end)))

  block_light_sections = Enum.chunk_every(chunk_array, 16, 16, 0)
  |> Enum.map(fn chunk_section ->
    chunk_section |> List.flatten() |> Enum.map(fn {_bs, _sl, bl} -> bl end)
  end)
  non_empty_block_light_sections = block_light_sections |> Enum.filter(fn section -> !Enum.all?(section, &(&1 == 0)) end)
  has_block_light = [false] ++ (block_light_sections |> Enum.map(fn section -> !Enum.all?(section, &(&1 == 0)) end)) ++ [false]
  empty_block_light = has_block_light |> Enum.map(&not/1)
  block_light_array = non_empty_block_light_sections |> Enum.map(&(Enum.reduce(&1, <<>>, fn light, acc ->
    <<acc::bitstring, light::size(4)>>
  end)))

  send(to, {:send_packet, %{
    packet_type: :chunk_data_and_update_light,
    chunk_x: cx, chunk_z: cz,
    heightmaps: heightmaps,
    data: data,
    block_entities: [],
    sky_light_mask: has_sky_light,
    block_light_mask: has_block_light,
    empty_sky_light_mask: empty_sky_light,
    empty_block_light_mask: empty_block_light,
    sky_light_arrays: sky_light_array |> Enum.map(&(%{sky_light_array: &1})),
    block_light_arrays: block_light_array |> Enum.map(&(%{block_light_array: &1}))
  }})
  :ok
  end

  defp long_aligned_bit_string_reduce(values, bpe, func) do
    values |> Enum.reduce(<<>>, fn value, acc ->
      ret = func.(value)
      next = <<acc::bitstring, ret::bitstring-size(bpe)>>
      # man i hope they dont suddenly change the size of a long
      if rem(bit_size(next), 64) + bpe > 64 do
        # gotta pad it
        <<next::bitstring, 0::big-size(64 - rem(bit_size(next), 64))>>
      else
        next
      end
    end)
  end

  @spec handle_packet(pos_integer(), binary(), %__MODULE__{}) :: [(%__MODULE__{} -> %__MODULE__{})]
  defp handle_packet(id, data, connstate, version, state) do
    try do
      packet = connstate.deserialize(id, version, data)
      case connstate.handle(packet, version, state) do
        :ok -> state
        {:error, reason} ->
          Logger.error("Error handling packet with ID #{inspect(id, base: :hex)} in state #{connstate}: #{reason}")
          send(self(), {:disconnect, "§cError handling packet #{inspect(id, base: :hex)}:\n#{reason}"})
          state
        newstate ->
          if is_map(newstate) do
            newstate
          else
            Logger.warning("State change to #{inspect(newstate)} is not a map! Did you forget to return :ok?")
            state
          end
      end
    rescue
      e ->
        if Application.get_env(:amethyst, :release, false) do
          send(self(), {:disconnect, "§cError handling packet #{inspect(id, base: :hex)}:\n#{Exception.format(:error, e, __STACKTRACE__)}"})
          Logger.error("Error handling packet with ID #{inspect(id, base: :hex)} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        else
          Logger.error("Error handling packet with ID #{inspect(id, base: :hex)} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        end
        state
    end
  end

  defp send_packet(socket, connstate, packet, version, state) do
    try do
      data = connstate.serialize(packet, version)
      container = if Map.get(state, :compression) == nil do
        # Packet ID is included in data
        Write.varint(byte_size(data)) <> data
      else
        threshold = Map.get(state, :compression, 0)
        data_length = byte_size(data)
        if data_length >= threshold do
          compressed = Write.varint(data_length) <> :zlib.compress(data)
          Write.varint(byte_size(compressed)) <> compressed
        else
          compressed = Write.varint(0) <> data
          Write.varint(byte_size(compressed)) <> compressed
        end
      end
      encrypted = if Map.get(state, :encryption_state) == nil do
        container
      else
        Map.get(state, :encryption_state) |> :crypto.crypto_update(container)
      end
      :gen_tcp.send(socket, encrypted)
    rescue
      e ->
        Logger.error("Error sending packet #{inspect(packet)} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        send(self(), {:disconnect, "§cError sending packet #{inspect(packet)}:\n#{Exception.format(:error, e, __STACKTRACE__)}"})
    end
  end

  defp disconnect(socket, reason, connstate, version, state) do
    Logger.info("Disconnecting connection #{inspect(socket)}")
    Logger.debug("Disconnecting connection #{inspect(socket)}: #{reason}")
    case connstate.disconnect(reason) do
      nil -> nil
      packet -> send_packet(socket, connstate, packet, version, state)
    end
    :gen_tcp.close(socket)
  end
end
